import { logger } from '../../lib/logger';
import { createWorkerBox } from '../../lib/workerbox';
import { resolvers } from './arg-resolvers';
import { replaceBaseUrl } from '../../lib/replace-base-url';
import { WebRuntimeConstants, WebSignalsRuntime, } from '@segment/analytics-signals-runtime';
import { getRuntimeCode } from '@segment/analytics-signals-runtime';
import { polyfills } from './polyfills';
import { loadScript } from '../../lib/load-script';
/**
 * Proxy around the analytics client
 */
class AnalyticsRuntime {
    constructor() {
        this.calls = {
            page: [],
            identify: [],
            track: [],
            alias: [],
            screen: [],
            group: [],
            reset: [],
        };
        // these methods need to be bound to the instance, rather than the prototype, in order to serialize correctly in the sandbox.
        this.track = (...args) => {
            try {
                // @ts-ignore
                const [eventName, props, options, cb] = resolvers.resolveArguments(
                // @ts-ignore
                ...args);
                this.calls.track.push([eventName, props, this.stamp(options), cb]);
            }
            catch (err) {
                // wrapping all methods in a try/catch because throwing an error won't cause the error to surface inside of workerboxjs
                console.error(err);
            }
        };
        this.identify = (...args) => {
            try {
                // @ts-ignore
                const [id, traits, options, cb] = resolvers.resolveUserArguments(...args);
                this.calls.identify.push([id, traits, this.stamp(options), cb]);
            }
            catch (err) {
                console.error(err);
            }
        };
        this.alias = (...args) => {
            try {
                const [userId, previousId, options, cb] = resolvers.resolveAliasArguments(
                // @ts-ignore
                ...args);
                this.calls.alias.push([userId, previousId, this.stamp(options), cb]);
            }
            catch (err) {
                console.error(err);
            }
        };
        this.group = (...args) => {
            try {
                // @ts-ignore
                const [id, traits, options, cb] = resolvers.resolveUserArguments(...args);
                this.calls.group.push([id, traits, this.stamp(options), cb]);
            }
            catch (err) {
                console.error(err);
            }
        };
        this.page = (...args) => {
            try {
                const [category, name, props, options, cb] = resolvers.resolvePageArguments(...args);
                this.stamp(options);
                this.calls.page.push([category, name, props, this.stamp(options), cb]);
            }
            catch (err) {
                console.error(err);
            }
        };
        this.screen = (...args) => {
            try {
                const [category, name, props, options, cb] = resolvers.resolvePageArguments(...args);
                this.stamp(options);
                this.calls.screen.push([category, name, props, this.stamp(options), cb]);
            }
            catch (err) {
                console.error(err);
            }
        };
        this.reset = () => {
            this.calls.reset.push([]);
        };
    }
    getCalls() {
        return this.calls;
    }
    /**
     * Stamp the context with the event origin to prevent infinite signal-event loops.
     */
    stamp(options) {
        if (!options) {
            options = {};
        }
        options.context = { ...options.context, __eventOrigin: { type: 'Signal' } };
        return options;
    }
}
class JavascriptSandbox {
    constructor() {
        this.workerbox = createWorkerBox();
    }
    async run(fn, scope) {
        try {
            const wb = await this.workerbox;
            await wb.run(fn, scope);
        }
        catch (err) {
            console.error('processSignal() error in sandbox', err, {
                fn,
            });
        }
    }
    async destroy() {
        const wb = await this.workerbox;
        await wb.destroy();
    }
}
export const normalizeEdgeFunctionURL = (functionHost, edgeFnDownloadURL) => {
    if (functionHost && edgeFnDownloadURL) {
        replaceBaseUrl(edgeFnDownloadURL, `https://${functionHost}`);
    }
    else {
        return edgeFnDownloadURL;
    }
};
const consoleWarnProcessSignal = () => console.warn('processSignal is not defined - have you set up auto-instrumentation on app.segment.com?');
export class IframeSandboxSettings {
    constructor(settings) {
        const fetch = settings.edgeFnFetchClient ?? globalThis.fetch;
        let processSignalNormalized = Promise.resolve(`globalThis.processSignal = function() {}`);
        if (settings.processSignal) {
            processSignalNormalized = Promise.resolve(settings.processSignal).then((str) => `globalThis.processSignal = ${str}`);
        }
        else if (settings.edgeFnDownloadURL) {
            processSignalNormalized = fetch(settings.edgeFnDownloadURL).then((res) => res.text());
        }
        else {
            consoleWarnProcessSignal();
        }
        this.processSignal = processSignalNormalized;
    }
}
export class WorkerSandbox {
    constructor(settings) {
        this.settings = settings;
        this.jsSandbox = new JavascriptSandbox();
    }
    async execute(signal, signals) {
        const analytics = new AnalyticsRuntime();
        const scope = {
            analytics,
        };
        logger.debug('processing signal', { signal, scope, signals });
        const code = [
            polyfills,
            await this.settings.processSignal,
            getRuntimeCode(),
            `signals.signalBuffer = ${JSON.stringify(signals)};`,
            'try { processSignal(' +
                JSON.stringify(signal) +
                ', { analytics, signals, SignalType, EventType, NavigationAction }); } catch(err) { console.error("Process signal failed.", err); }',
        ].join('\n');
        await this.jsSandbox.run(code, scope);
        const calls = analytics.getCalls();
        return calls;
    }
    destroy() {
        void this.jsSandbox.destroy();
    }
}
// ProcessSignal unfortunately uses globals. This should change.
// For now, we are setting up the globals between each invocation
const processWithGlobalScopeExecutionEnv = (signal, signalBuffer) => {
    const g = globalThis;
    const processSignal = g['processSignal'];
    if (typeof processSignal == 'undefined') {
        consoleWarnProcessSignal();
        return undefined;
    }
    // Load all constants into the global scope
    Object.entries(WebRuntimeConstants).forEach(([key, value]) => {
        g[key] = value;
    });
    // processSignal expects a global called `signals` -- of course, there can local variable naming conflict on the client, which is why globals were a bad idea.
    const analytics = new AnalyticsRuntime();
    const signals = new WebSignalsRuntime(signalBuffer);
    const originalAnalytics = g.analytics;
    if (originalAnalytics instanceof AnalyticsRuntime) {
        throw new Error('Invariant: analytics variable was not properly restored on the previous execution. This indicates a concurrency bug');
    }
    const originalSignals = g.signals;
    try {
        g['analytics'] = analytics;
        g['signals'] = signals;
        processSignal(signal, {
            // we eventually want to get rid of globals and processSignal just uses local variables.
            // TODO: update processSignal generator to accept params like these for web (mobile currently uses globals for their architecture -- can be changed but hard).
            analytics: analytics,
            signals: signals,
            // constants
            EventType: WebRuntimeConstants.EventType,
            NavigationAction: WebRuntimeConstants.NavigationAction,
            SignalType: WebRuntimeConstants.SignalType,
        });
    }
    finally {
        // restore globals
        g['analytics'] = originalAnalytics;
        g['signals'] = originalSignals;
    }
    return analytics.getCalls();
};
export class GlobalScopeSandbox {
    constructor(settings) {
        logger.debug('Initializing global scope sandbox');
        this.htmlScriptLoaded = loadScript(settings.edgeFnDownloadURL);
    }
    async execute(signal, signals) {
        await this.htmlScriptLoaded;
        return processWithGlobalScopeExecutionEnv(signal, signals);
    }
    destroy() { }
}
export class NoopSandbox {
    execute(_signal, _signals) {
        return Promise.resolve(undefined);
    }
    destroy() { }
}
//# sourceMappingURL=sandbox.js.map